#! /usr/bin/env python

import os
import sys
import math
import random

import pygame
from retrogamelib import display
from retrogamelib import button
from retrogamelib import clock
from retrogamelib import util
from retrogamelib import font
from retrogamelib import gameobject
from retrogamelib import dialog
from retrogamelib.constants import *

util.set_global_sound_volume(0.25)

class Player(gameobject.Object):
    
    def __init__(self, y, ai=False, speed=5, diff=1):
        gameobject.Object.__init__(self, self.groups)
        facing = "up"
        if y < 120:
            facing = "down"
        self.images = [
            util.load_image("data/bat-%s-1.png" % facing),
            util.load_image("data/bat-%s-2.png" % facing),
            util.load_image("data/bat-%s-3.png" % facing),
        ]
        self.image = self.images[0]
        self.rect = pygame.Rect(0, 0, 32, 32)
        self.rect.midbottom = (128, y)
        self.swinging = 0
        self.aim = pygame.Rect(0, 0, 2, 2)
        self.speed = speed
        self.diff = diff
        self.x = self.rect.centerx
        
        # On top or bottom?
        if self.rect.centery > 120:
            self.aim.top = 48
            self.offset = -12
        else:
            self.offset = 0
            self.aim.bottom = 240-48
        
        self.ai = ai
        self.serving = False

    def do_ai(self, ball, player):
        
        # If the ball touches us, swing!
        if ball.rect.colliderect(self.rect):
            self.swinging = 7
            
            self.aim.centerx = self.rect.centerx
            if self.diff > 1:
                bounds = 200
            
            if not random.randrange(3) and self.diff == 3:
                if player.rect.centerx > 120:
                    self.aim.centerx -= 16
                else:
                    self.aim.centerx += 16
            if not random.randrange(2) and self.diff == 2:
                if player.rect.centerx > 120:
                    self.aim.centerx -= 16
                else:
                    self.aim.centerx += 16
            self.aim.centerx += random.randrange(-12, 12)
       
        # Follow the ball
        diff = self.rect.centerx-ball.rect.centerx
        if diff > self.speed:
            diff = self.speed
        if diff < -self.speed:
            diff = -self.speed
        self.x -= diff

    def update(self):
        
        # For layering
        self.z = self.rect.bottom
        self.rect.centerx = self.x
        
        # Make sure we're not some stupid computer player
        if not self.ai:
            
            # Swing the bat
            if button.is_pressed(A_BUTTON) and self.swinging <= 0:
                self.swinging = 7
                self.aim.centerx = self.rect.centerx
                self.serving = False
            
            # Move and aim the bat
            if not self.serving:
                if button.is_held(LEFT):
                    if self.swinging > 0: # aim the shot
                        self.aim.x -= 24 + (self.diff-1)*6
                    self.x -= self.speed # move the bat
                if button.is_held(RIGHT):
                    if self.swinging > 0: # aim the shot
                        self.aim.x += 24 + (self.diff-1)*6
                    self.x += self.speed # move the bat
        
        # Animation!
        self.swinging -= 1
        if self.swinging > 0:
            if self.swinging > 0:
                self.image = self.images[2]
            if self.swinging > 5:
                self.image = self.images[1]
        else:
            self.image = self.images[0]
       
        # Keep the bat on the screen
        if self.x < 0:
            self.x = 0
        if self.x > 256-32:
            self.x = 256-32

    def draw(self, surface):
        surface.blit(self.image, (self.rect.x, self.rect.y+self.offset))

class Ball(gameobject.Object):
    
    def __init__(self, dy=9, bounce=8, float=1.6):
        gameobject.Object.__init__(self, self.groups)
        self.image = util.load_image("data/ball.png")
        self.shadow = util.load_image("data/shadow.png")
        self.rect = pygame.Rect(0, 0, 3, 3)
        self.rect.center = (128, 200)
       
        self.dx = 0
        self.dy = dy
        self.bounce = bounce
        self.bounce_speed = 0
        self.height = 22
        self.float = float
      
        self.surface = "table"
        self.bottom = 0
      
        self.served = False
        self.out = False
        self.net = False
        self.bounces = 0
        self.winner = False

    def update(self):
        
        # For layering
        self.z = self.rect.bottom
        
        # Only move if the ball's been served
        if self.served:
            self.rect.x += self.dx
            self.rect.y += self.dy
            self.height += self.bounce_speed
        
        # Make the bounce velocity decrease, aka apply gravity
        self.bounce_speed -= self.float
       
        # How high are we?
        if self.surface == "table":
            self.bottom = 0
        elif self.surface == "floor":
            self.bottom = -12
        
        # Did it bounce?
        if self.height < self.bottom:
            self.bounce_speed = self.bounce
            self.height = 0
            self.bounces += 1
            util.play_sound("data/bounce.ogg")
            #ShotSpot((self.rect.centerx, self.rect.centery+self.bottom))
           
            # Did it land out?
            if self.surface == "floor":
                self.out = True
        
        # Is it a winner?
        if self.rect.top < 0 or self.rect.bottom > 240 or \
            self.bounces > 1:
            self.winner = True

    def draw(self, surface):
        surface.blit(self.shadow, (self.rect.x-3, self.rect.y-self.bottom-3))
        surface.blit(self.image, (self.rect.x-3, self.rect.y-self.height-3))

    def hit(self, player):
        
        util.play_sound("data/hit.ogg")
        
        # Reset the number of times the ball has bounced
        self.bounces = 0
        
        # Record the last player that's hit the ball
        self.last_player = player
        
        # We've been hit! The ball is served.
        if not self.served:
            self.served = True
       
        # Reverse y speed
        if player.rect.centery > 120:
            if self.dy < 0: # Already hit the ball
                return
            self.dy = -abs(self.dy)
        else:
            if self.dy > 0: # Already hit the ball
                return
            self.dy = abs(self.dy)
        
        # Bounce off the paddle
        self.bounce_speed = 8
        
        # Make sure we don't hit the ball too low.
        if self.height < 5:
            self.height = 5
        
        # Move away from the paddle on impact
        self.rect.y += self.dy*2
        
        # Calculate x speed
        offset = abs(self.height/5)
        xdiff = player.aim.centerx - player.rect.centerx
        xdiff /= 8.0
        self.dx = xdiff
        if self.dx < -6:
            self.dx = -6
        if self.dx > 6:
            self.dx = 6

class Net(gameobject.Object):
    
    def __init__(self):
        gameobject.Object.__init__(self, self.groups)
        self.image = util.load_image("data/net.png")
        self.z = 120
        self.rect = self.image.get_rect(bottomleft=(32, 120))
    
    def draw(self, surface):
        surface.blit(self.image, self.rect)

class ShotSpot(gameobject.Object):
    
    def __init__(self, pos):
        gameobject.Object.__init__(self, self.groups)
        self.images = [
            util.load_image("data/bounce-1.png"),
            util.load_image("data/bounce-2.png"),
            util.load_image("data/bounce-3.png"),
        ]
        self.image = self.images[0]
        self.z = -1
        self.rect = self.image.get_rect(center=pos)
        self.frame = 0
    
    def draw(self, surface):
        self.image = self.images[self.frame/2%3]
        self.frame += 1
        if self.frame >= 5:
            self.kill()
        surface.blit(self.image, self.rect)

class Menu(object):
    
    def __init__(self):
        self.font = font.Font(NES_FONT, (255, 255, 255))
        self.menu = dialog.Menu(self.font, [" Level 1 ", " Level 2 ", " Level 3 "])
        self.background = util.load_image("data/menu.png")
        
        while 1:
            button.handle_input()
            
            if button.is_pressed(UP):
                self.menu.move_cursor(-1)
                util.play_sound("data/select.ogg")
            if button.is_pressed(DOWN):
                self.menu.move_cursor(1)
                util.play_sound("data/select.ogg")
            if button.is_pressed(START):
                util.play_sound("data/enter.ogg")
                game = Game(self.menu.get_option()[0]+1)
            
            screen = display.get_surface()
            screen.blit(self.background, (0, 0))
            self.menu.draw(screen, (85, 184))
            ren = self.font.render("retrogamelib")
            screen.blit(ren, (128-ren.get_width()/2, 80))
            display.update()

class Game(object):
    
    def __init__(self, level):
        self.objects = gameobject.Group()
        self.players = gameobject.Group()
        self.balls = gameobject.Group()
        Player.groups = [self.objects, self.players]
        Ball.groups = [self.objects, self.balls]
        Net.groups = [self.objects]
        ShotSpot.groups = [self.objects]
        
        self.player1 = Player(240-32, diff=level)
        self.player1.serving = True
        
        if level == 1:
            speed = 5
        if level == 2:
            speed = 5.25
        if level == 3:
            speed = 5.1
        
        self.player2 = Player(32, True, speed, level)
        
        self.bounce = 8
        self.float = 1.6
        if level == 1:
            self.bspeed = 7
            self.float = 1.4
        if level == 2:
            self.bspeed = 8
        if level == 3:
            self.bspeed = 10
            self.float = 1.9
        
        self.ball = Ball(self.bspeed, self.bounce, self.float)
        self.net = Net()
        self.table = pygame.Rect(34, 32, 256-64, 240-64)
        
        self.background = util.load_image("data/background.png")
        self.font = font.Font(NES_FONT, (255, 255, 255))
        self.blackfont = font.Font(NES_FONT, (1, 1, 1))
        
        self.point_over = False
        self.p1score = 0
        self.p2score = 0
        self.game_finished = False
        
        self.message = ""
        self.msgtimer = 0
        self.paused = False
        
        self.loop()
    
    def show_message(self, message, msgtimer=25):
        self.message = message
        self.msgtimer = msgtimer
    
    def new_ball(self):
        self.ball.kill()
        self.ball = Ball(self.bspeed, self.bounce, self.float)
        self.player1.x = self.ball.rect.centerx
        self.player2.x = self.ball.rect.centerx
        self.player1.serving = True
        self.point_over = False
    
    def stop_ball(self):
        self.point_over = True
        for p in self.players:
            p.hitting = 0
    
    def update(self):
        
        # Tick, tock, tick, tock
        clock.tick()
        
        if not self.paused:
            # Update the objects
            for object in self.objects:
                if isinstance(object, Player):
                    object.update()
                else:
                    object.update()
            
            # Do ai for player 2
            self.player2.do_ai(self.ball, self.player1)
       
        # Did someone hit the ball?
        if not self.point_over:
            for player in self.players:
                if player.swinging > 3:
                    if player.rect.centery > 120:
                        rect = pygame.Rect(self.ball.rect)
                        rect.y -= self.ball.height
                    else:
                        rect = self.ball.rect
                    if player.rect.colliderect(rect):
                        self.ball.hit(player)
        
        # What surface is the ball on?
        if self.ball.rect.colliderect(self.table):
            self.ball.surface = "table"
        else:
            self.ball.surface = "floor"
        
        # Hit the net?
        if self.ball.rect.colliderect(self.net.rect) and \
            self.ball.rect.bottom > self.net.rect.bottom-8:
            if self.ball.height < 22:
                if not self.ball.net:
                    if self.ball.dy < 0:
                        self.ball.dy = 4
                    elif self.ball.dy > 0:
                        self.ball.dy = -4
                    self.ball.rect.y += self.ball.dy
                self.ball.dx *= 0.5
                self.ball.net = True
                if not self.point_over:
                    self.show_message("net!")
                    util.play_sound("data/net.ogg")
                    self.stop_ball()
                    if self.ball.last_player == self.player1:
                        self.p2score += 1
                    else:
                        self.p1score += 1
        
        if not self.point_over:
            
            # Did you hit the ball out?
            if self.ball.out or self.ball.net:
                if self.ball.last_player == self.player1:
                    self.p2score += 1
                else:
                    self.p1score += 1
                self.stop_ball()
                self.show_message("out!")
                util.play_sound("data/out.ogg")
            
            # Did you hit a winner?
            if self.ball.winner:
                if self.ball.last_player == self.player1:
                    self.p1score += 1
                else:
                    self.p2score += 1
                self.stop_ball()
                self.show_message("winner!")
                util.play_sound("data/winner.ogg")
    
        # At the end of each message, spawn a new ball
        self.msgtimer -= 1
        if self.point_over:
            if self.msgtimer < 0:
                self.new_ball()
        
        # Who won?
        if self.p1score >= 11:
            self.p1score = 11
            if not self.game_finished:
                self.show_message("you win!", 50)
                self.game_finished = True
        if self.p2score >= 11:
            self.p2score = 11
            if not self.game_finished:
                self.show_message("you lose!", 50)
                self.game_finished = True
        if self.game_finished and self.msgtimer <= 0:
            self.running = False
    
    def handle_input(self):
        button.handle_input()
        if button.is_pressed(START):
            self.paused ^= 1
        if self.paused:
            if button.is_pressed(SELECT):
                self.running = False
    
    def draw(self):
        screen = display.get_surface()
        screen.blit(self.background, (0, 0))
        
        # Draw all the objects
        for object in self.objects:
            object.draw(screen)
       
        # Human's scoreboard
        pygame.draw.rect(screen, (0, 0, 0), (1, 90, 30, 24))
        pygame.draw.rect(screen, (255, 255, 255), (1, 90, 30, 24), 1)
        ren = self.font.render("you")
        screen.blit(ren, (4, 96))
        ren = self.font.render("%d" % self.p1score)
        screen.blit(ren, (12, 104))
        
        # Computer's scoreboard
        pygame.draw.rect(screen, (0, 0, 0), (225, 90, 30, 24))
        pygame.draw.rect(screen, (255, 255, 255), (225, 90, 30, 24), 1)
        ren = self.font.render("com")
        screen.blit(ren, (230, 96))
        ren = self.font.render("%d" % self.p2score)
        screen.blit(ren, (238, 104))
        
        if self.msgtimer > 0:
            ren = self.blackfont.render(self.message)
            screen.blit(ren, (129-ren.get_width()/2, 
                121-ren.get_height()/2))
            ren = self.font.render(self.message)
            screen.blit(ren, (128-ren.get_width()/2, 
                120-ren.get_height()/2))
        if self.paused:
            ren = self.blackfont.render("Press select to quit")
            screen.blit(ren, (129-ren.get_width()/2, 
                121-ren.get_height()/2))
            ren = self.font.render("Press select to quit")
            screen.blit(ren, (128-ren.get_width()/2, 
                120-ren.get_height()/2))
        
        display.update()
    
    def loop(self):
        self.running = True
        while self.running:
            self.update()
            self.handle_input()
            self.draw()

def main():
    display.init(2.0, "Ping Pong", NESRES)
    util.play_sound("data/intro.ogg")
    menu = Menu()

if __name__ == "__main__":
    main()
